/*
 * socks4 specific negotiation
 * marius aamodt eriksen (marius@umich.edu)
 */

#include "nylon.h"

#define SOCKS4_MAX_USERID 1024
#define SOCKS4_CD_CONNECT 1
#define SOCKS4_CD_BIND 2
#define SOCKS4_CD_GRANT 90
#define SOCKS4_CD_REJECT 91

struct socks4_hdr {
	u_char vn;
	u_char cd;
	u_int16_t d_p;     
	u_int32_t d_addr;  
};

int socks4_connect __P((struct conf*, int, struct sockaddr_in*,
						struct socks4_hdr*));
int socks4_bind __P((struct conf*, int, struct sockaddr_in*,
					 struct socks4_hdr*));

int
negotiate4(struct conf *c, int client_s, int session_no)
{
	u_char *userid;
	u_int32_t i, j;
	int target_s, ret;
	struct socks4_hdr hdr4;
	struct sockaddr_in target_in, connect_in;

	printv(2, "entering socks4 specific negotiation\n");
	
	userid = (unsigned char*)malloc(DEFAULT_MALLOC);

	i = 0;
	j = 1;
	ret = -ERROR_UNDEF;
	hdr4.vn = 4;
	
	/* getting the seven first bytes after version, until USERID */
	if (recvn(client_s, &hdr4.cd, 7, 0) == -1) {
		ret = -ERROR_RECV;
		goto cleanup;
	}
	
	while (recvn(client_s, userid+i, 1, 0) != -1)
		if ((userid[i]) == 0) {
			userid[i] = '\0';
			break;
		} else {
			if (++i % DEFAULT_MALLOC == 0)
				userid = (unsigned char*)realloc(userid, ++j *
												 DEFAULT_MALLOC);
			if (j * DEFAULT_MALLOC > SOCKS4_MAX_USERID) {
				printv(2, "header too long, bailing out");
				goto cleanup;
			}
		}
	
	memset(&target_in, 0, sizeof(target_in));
	target_in.sin_family = AF_INET;
	target_in.sin_port = hdr4.d_p;               /* already in network order */
	target_in.sin_addr.s_addr = hdr4.d_addr;     /* ditto                    */

	switch (hdr4.cd) {
	case SOCKS4_CD_CONNECT:
		target_s = socks4_connect(c, client_s, &target_in, &hdr4);
	default:
		ret = -ERROR_UNSUP;
		goto cleanup;
	}
		
	/* this will pass through errors, XXX handle where?? */

	ret = target_s;

 cleanup:
	free(userid);
	return ret;
}

int
socks4_connect(struct conf *c, int client_s, struct sockaddr_in *target_in,
			   struct socks4_hdr *hdr4)
{
	int ret, target_s;
	struct sockaddr_in connect_in;
	
	ret = -ERROR_UNDEF;

	if ((target_s = socket(AF_INET, SOCK_STREAM, 0)) < 0) 
		return -ERROR_SOCKET;

	if (c->if_conn) {
#ifdef WIN32
		struct hostent *cinfhost;
		cinfhost = gethostbyname(c->if_conn);
		connect_in.sin_addr = *(struct in_addr*)cinfhost->h_addr;
#else /* WIN32 */
		connect_in.sin_addr = get_ip_from_if(c->if_conn);
#endif /* WIN32 */
		connect_in.sin_port = 0;
		if (bind(target_s, (struct sockaddr*) &connect_in,
				 sizeof(connect_in)) < 0) {
			ret = -ERROR_BIND;
			goto cleanup;
		}
	}

	if (connect(target_s, (struct sockaddr*)&target_in, sizeof(target_in)) < 0)
		hdr4->cd = SOCKS4_CD_REJECT;
	else
		hdr4->cd = SOCKS4_CD_GRANT;

	hdr4->vn = 0; 

	/*
	 * XXX - "irregular" architecures may not like this; extra padding
	 * on the structs
	 */
	if (sendn(client_s, hdr4, sizeof(*hdr4), 0) < 0) {
		ret = -ERROR_REPLY;
		goto cleanup;
	}

	if (hdr4->cd == SOCKS4_CD_REJECT) {
		ret = -ERROR_CONNECT;
		goto cleanup;
	}

	return target_s;

 cleanup:
	close(target_s);
	return ret;
}

#if 0
int
socks4_bind(struct conf *c, int client_s, struct sockaddr_in *target_in,
			struct socks4_hdr *hdr4)
{

}
#endif /* 0 */

/* copyright (c) 2001  
 * marius aamodt eriksen
 * all rights reserved
 * 
 * permission is granted to use, copy, create derivative works and
 * redistribute this software and such derivative works for any purpose,
 * so long as the name of the author is not used in any advertising or 
 * publicity pertaining to the use or distribution of this software 
 * without specific, written prior authorization.  if the above copyright 
 * notice or any other identification of the author is included in any 
 * copy of any portion of this software, then the disclaimer below must 
 * also be included.
 * 
 * this software is provided as is, without representation from the
 * the author as to its fitness for any purpose, and without warranty by 
 * the author of any kind, either express or implied, including without 
 * limitation the implied warranties of merchantability and fitness for 
 * a particular purpose.  the author shall not be liable for any damages,
 * including special, indirect, incidental, or consequential damages,
 * with respect to any claim arising out or in connection with the use
 * of the software, even if it has been or is hereafter advised of the
 * possibility of such damages. */
